#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>

#include "cgprof_main.h"

#include "fptype.h"
#include "fpblas.h"
#include "task_log.h"
#include "matfprint.h"

#include "cgas_workspace.h"
#include "as_man.h"
#include "as_man.h"
#include "async_struct.h"
#include "selfsched.h"

#include "bblas_dot.h"
#include "bblas_dot_async.h" 
#include "bblas_gemm.h"
#include "bblas_axpy.h" 
#include "bblas_axpy_async.h" 
#include "bblas_copy.h"


typedef enum {
	DOT_ALPHA1, 
	DOT_ALPHA2,
	DOT_ALPHA3, 
} dot_t;


#if USE_SPARSE
static const char cgid[] = "cgprofs";
#else
static const char cgid[] = "cgprof";
#endif


/* The first CP preferably finishes before the first GEMM executes, so that we have a good profile of p. The asynchronous DOT
should be preferred over the GEMM, because we want to overlap the DOTs with the GEMMs, and complete as much of the dot product as possible. 
The subsequent AXPY must have lowest priority, because we only want to use an imprecise dot product if it improves performance. */
#define PRIOR_FIRSTCP		9999999
#define PRIOR_GEMM			99999
#define PRIOR_ASYNCDOT		999999
#define PRIOR_ASYNCAXPY		9999


int CGPROF(int bm, int bn, int n, void *A, int s, void *b, void *xbuff, int *offs, double tol, int steps, void *work, unsigned long works, int lookahead, int async, double profile, int warmup, int cglog, int release, double orth_fac) 
{
	typedef struct { 
		async_t sync[3];
		fp_t buff;
	} worklay_t;

	fp_t *lxbuff = xbuff;
	fp_t *x[4] = {lxbuff, lxbuff+n, lxbuff+2*n, lxbuff+3*n};
	fp_t **r; 
	fp_t **tmp;  
	fp_t **p; 
	fp_t **alpha1;
	fp_t **alpha2;
	worklay_t *worklay = (worklay_t*) work;
	fp_t *orth = malloc(s*s*sizeof(fp_t));
	fp_t porth = INFINITY;

	/* allocate work space */
	int dupl = 4;
	if ( cgas_malloc(n, s, dupl, &worklay->buff, works, &p, &tmp, &r, &alpha1, &alpha2) ) {
		return -1;
	}

	/* initializations */
	async_t *sync =  async_init(worklay->sync, 3, n, bm, -2, 1);
	async_profile(&sync[DOT_ALPHA2], profile);
	asman_t *asman = asman_init(4);
	if ( cglog ) {
		async_log(&sync[DOT_ALPHA1], 256, "cgprof");
		async_log(&sync[DOT_ALPHA2], 1024, "cgprof");
		async_log(&sync[DOT_ALPHA3], 1024, "cgprof");
	}

  	int i = asman_nexti(asman);
	
	/* r[0] = b - A x[0] */
	BBLAS_COPY(PRIOR_FIRSTCP, bm, bn, n, s, b, r[i]);	                     						
	BBLAS_GEMM(PRIOR_FIRSTCP, OMPSSBLAS_NTRANSP, OMPSSBLAS_NTRANSP, bm, bn, bm, n, s, n, FP_MONE, A, n, x[i], n, FP_ONE, r[i], n); 					

	/* p[0] = r[0] */
	BBLAS_COPY(PRIOR_FIRSTCP, bm, bn, n, s, r[i], p[i]); 				
	/* alpha1[0] = <r[0], r[0]> */
	BBLAS_DOT(-1, &sync[DOT_ALPHA1], bm, bn, n, s, r[i], r[i], alpha1[i]);            	

	int iprev = 0;
	int aheadc = 0;
	async_stat_t status = STAT_SYNC;

	asman_update(asman, 0, 0);

	int k; 
	for ( k=0; k<steps; ++k ) {
		i = asman_nexti(asman);
		printf("CGPROF k: %d\n", k);

		if ( status == STAT_AHEAD ) {
			int prevk = k - 1;
			status = BBLAS_GEMM_PROFPRED(k, &sync[DOT_ALPHA2], &sync[DOT_ALPHA3], PRIOR_GEMM, bm, bn, bm, n, s, n, FP_ONE, A, p[iprev], FP_NOUGHT, tmp[i],\
							prevk, iprev, asman, &sync[DOT_ALPHA1], alpha1, tol);

			if ( status == STAT_AHEAD ) {
				#pragma omp taskwait on (*alpha1[iprev])
				status = ASMAN_BREAK(prevk, iprev, asman, alpha1, &sync[DOT_ALPHA1], 1, tol);
			}
			if ( status == STAT_BROKEN || status == STAT_CONVERGED ) {
					break;
			}
		} else {
			BBLAS_GEMM_PRIOR(k, &sync[DOT_ALPHA2], &sync[DOT_ALPHA3], PRIOR_GEMM, bm, bn, bm, n, s, n, FP_ONE, A, p[iprev], FP_NOUGHT, tmp[i]);
		}

		if ( async && warmup<=0 ) {
			if ( release == 1989 ) {
				BBLAS_DOT_SCHED_ASYNC(k, &sync[DOT_ALPHA2], PRIOR_ASYNCDOT, bm, bn, n, s, tmp[i], p[iprev], alpha2[i], release); 		
				// alpha2[i] = <tmp[i], p[i]> (partial)
				BBLAS_SCAL_CPAXPY_COMB_ASYNC(k, &sync[DOT_ALPHA2], PRIOR_ASYNCAXPY, bm, bn, n, s, -1.0, alpha1[iprev], alpha2[i], tmp[i], p[iprev], r[iprev], x[iprev], r[i], x[i]);    	// r -=  alpha^T * tmp[i]
			} else {
				BBLAS_DOT_SCHED_ASYNC_RELEASE(k, &sync[DOT_ALPHA2], PRIOR_ASYNCDOT, bm, bn, n, s, tmp[i], p[iprev], alpha2[i], release); 		
				// alpha2[i] = <tmp[i], p[i]> (partial)
				BBLAS_SCAL_CPAXPY_COMB_ASYNC_RELEASE(k, &sync[DOT_ALPHA2], PRIOR_ASYNCAXPY, bm, bn, n, s, -1.0, alpha1[iprev], alpha2[i], tmp[i], p[iprev], r[iprev], x[iprev], r[i], x[i]);    	// r -=  alpha^T * tmp[i]
			}
		} else {
			BBLAS_DOT(k, &sync[DOT_ALPHA2], bm, bn, n, s, tmp[i], p[iprev], alpha2[i]); 		// alpha2[i] = <tmp[i], p[i]> (partial)
			BBLAS_CPAXPY_COMB(bm, bn, n, s, -1.0, alpha1[iprev], alpha2[i], tmp[i], p[iprev], r[iprev], x[iprev], r[i], x[i]);    	// r -=  alpha^T * tmp[i]
		}

		/* alpha1[i+1] = <r, r]> */
		BBLAS_DOT(k, &sync[DOT_ALPHA1], bm, bn, n, s, r[i], r[i], alpha1[i]);             					
		/* p[i+1] = r[i+1] + transpose(beta[i]) * p[i] */
		BBLAS_EXTM_AXPY(PRIOR_ASYNCDOT, bm, bn, n, s, alpha1[i], alpha1[iprev], p[iprev], r[i], p[i]); 	

		if ( ! lookahead ) {
			#pragma omp taskwait
		}

		BLAS_gemm(OMPSSBLAS_TRANSP, OMPSSBLAS_NTRANSP, s, s, n, FP_ONE, p[i], n, tmp[i], n, FP_NOUGHT, orth, s);
		*orth = FP_ABS(*orth);
		printf("orth: %e\n", *orth);
		if( async && warmup <= 0 ) {
			if (isgreater(*orth, porth * orth_fac)){
				printf("P orthogonality threashold failed: %e > %e\n", *orth, porth);
				break;
			}
		}

		prof_dump(&(sync[DOT_ALPHA2].prof));
		porth = *orth;

		status = ASMAN_BREAK(k, i, asman, alpha1, &sync[DOT_ALPHA1], 0, tol);
		if ( status == STAT_BROKEN || status == STAT_CONVERGED ) {
			printf("breaking\n");
			break;	
		} 

		iprev = i;
		aheadc += status == STAT_AHEAD ? 1 : 0;
		--warmup;
	}

	if ( status == STAT_AHEAD ) {
		#pragma omp taskwait on (*alpha1[iprev])
		ASMAN_BREAK(k-1, i, asman, alpha1, &sync[DOT_ALPHA1], 1, tol);
	}

	*offs = asman_best(asman);
	asman_fini(asman);

#pragma omp taskwait

	//printf("CGsched: aheadc %i\n", aheadc);

	free(p);

	return k;
}
